/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tmakeGraph.cc                                */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "tmakeGraph.h"

/* Warning: using this code assumes that your compiler supports partial
   expression evaluation (e.g., in (A || B), if A is true, B is not
   evaluated at all) */

/* This function swaps data about file names, contents and types
   inside a given cmdLineParams structure */
void dynaSwap(cmdLineParams &params, int index1, int index2) {
   char *tempString = new char[strlen(params.fileName[index1]) + 1];
   int tempContents = params.fileContents[index1];
   file_type tempType = params.fileType[index1];
   ASnode tempPrepend = params.prependAS[index1];
   strcpy (tempString, params.fileName[index1]);
   delete [] params.fileName[index1];
   params.fileName[index1] = new char[strlen(params.fileName[index2]) + 1];
   strcpy (params.fileName[index1], params.fileName[index2]);
   delete [] params.fileName[index2];
   params.fileName[index2] = new char[strlen(tempString) + 1];
   strcpy (params.fileName[index2], tempString);
   delete [] tempString;
   params.fileContents[index1] = params.fileContents[index2];
   params.fileContents[index2] = tempContents;
   params.fileType[index1] = params.fileType[index2];
   params.fileType[index2] = tempType;
   params.prependAS[index1] = params.prependAS[index2];
   params.prependAS[index2] = tempPrepend;
}

/* This function parses command line parameters and writes them
   inside a structure */
bool parseArguments (int argc, char **argv, cmdLineParams &params, char *error) {
   int processedContents, curIndex;
   long int value;
   bool expectComment, parsed, expectFile, typeSet, fileList, expectOutFile, expectVpoint,
        vpoint, expectGraphFile, commentSet;
   char *err;
   params.verbose = false;
   params.showHelp = false;
   params.force = false;
   params.nonBGPfiles = params.BGPfiles = 0;
   processedContents = 0;
   curIndex = 0;
   expectComment = expectFile = expectOutFile = expectGraphFile =
      expectVpoint = vpoint = commentSet = false;
   fileList = false;
   strcpy (error, "");
   for (int i = 1; i <= argc - 1; i++) {
      parsed = false;
      if (!parsed && expectVpoint) {
         value = strtol(argv[i], &err, 10);
         if (value < 1 || value > 65535 || (strlen(err) > 0)) {
            strcat (error, "Invalid vantage point number: `");
            strcat (error, argv[i]);
            strcat (error, "'");
            return false;
         }
         params.prependAS[curIndex] = value;
         expectVpoint = false;
         expectFile = true;
         parsed = true;
      }
      if (!parsed && expectFile) {
         // This must be a file name
         if (params.fileType[curIndex] == FILETYPE_BGPDUMP &&
             (params.fileContents[curIndex] & ~FILECONTENTS_PREFIXCOVERING)) {
            strcat (error, "BGP dumps can only be used to get IP address covering");
            return false;
         }
         if (params.fileType[curIndex] == FILETYPE_VANTAGE &&
             (params.fileContents[curIndex] & FILECONTENTS_PATHCOVERING ||
              params.fileContents[curIndex] & FILECONTENTS_PREFIXCOVERING)) {
            strcat (error, "Vantage files do not contain covering information");
            return false;
         }
         if (params.fileType[curIndex] == FILETYPE_PATH &&
             (params.fileContents[curIndex] & FILECONTENTS_PREFIXCOVERING)) {
            strcat (error, "Cannot get IP address covering from a path file");
            return false;
         }
         if (params.fileType[curIndex] == FILETYPE_UNKNOWN &&
             (params.fileContents[curIndex] & ~(FILECONTENTS_NODES | FILECONTENTS_EDGES))) {
            strcat (error, "Unoriented edges can only be used to get nodes and/or edges");
            return false;
         }
         if (params.fileType[curIndex] == FILETYPE_BGPDUMP) {
            fileList = true;
         }
         params.fileName[curIndex] = new char[strlen(argv[i]) + 1];
         strcpy (params.fileName[curIndex], argv[i]);
         processedContents |= params.fileContents[curIndex];
         curIndex++;
         expectFile = false;
         parsed = true;
      }
      if (expectGraphFile) {
         // This must be the name of a source graph
         if (params.graphFile != NULL) {
            strcat (error, "There is more than one input graph");
            return false;
         }
         params.graphFile = new char[strlen(argv[i]) + 1];
         strcpy (params.graphFile, argv[i]);
         expectGraphFile = false;
         parsed = true;
      }
      if (expectOutFile) {
         // This must be the output graph file name
         if (params.outFile != NULL) {
            strcat (error, "There is more than one output file");
            return false;
         }
         params.outFile = new char[strlen(argv[i]) + 1];
         strcpy (params.outFile, argv[i]);
         expectOutFile = false;
         parsed = true;
      }
      if (expectComment) {
         if (commentSet) {
            strcat (error, "There is more than one comment");
            return false;
         }
         strncpy(params.comment, argv[i], MAX_COMMENT_CHARS);
         // Terminate comment string
         params.comment[MAX_COMMENT_CHARS] = '\0';
         commentSet = true;
         expectComment = false;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--comment") == 0) {
         expectComment = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--enhance") == 0) {
         expectGraphFile = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-f") == 0) {
         params.force = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-v") == 0) {
         params.verbose = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-o") == 0) {
         expectOutFile = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--help") == 0) {
         params.showHelp = true;
         parsed = true;
      }
      if (!parsed && strncmp(argv[i], "--file:", 7) == 0) {
         // This might be an input file
         params.fileContents[curIndex] = 0;
         typeSet = false;
         for (unsigned int i2 = 7; i2 <= strlen(argv[i]) - 1; i2++) {
            switch (argv[i][i2]) {
               case 'n':   if (params.fileContents[curIndex] & FILECONTENTS_NODES) {
                              strcat (error, "Duplicate use of `n' flag for file contents");
                              return false;
                           }
                           params.fileContents[curIndex] |= FILECONTENTS_NODES;
                           expectFile = true;
                           break;
               case 'e':   if (params.fileContents[curIndex] & FILECONTENTS_EDGES) {
                              strcat (error, "Duplicate use of `e' flag for file contents");
                              return false;
                           }
                           params.fileContents[curIndex] |= FILECONTENTS_EDGES;
                           expectFile = true;
                           break;
               case 'd':   if (params.fileContents[curIndex] & FILECONTENTS_EDGEDIRECTION) {
                              strcat (error, "Duplicate use of `o' flag for file contents");
                              return false;
                           }
                           params.fileContents[curIndex] |= FILECONTENTS_EDGEDIRECTION;
                           expectFile = true;
                           break;
               case 'c':   if (params.fileContents[curIndex] & FILECONTENTS_PATHCOVERING) {
                              strcat (error, "Duplicate use of `p' flag for file contents");
                              return false;
                           }
                           params.fileContents[curIndex] |= FILECONTENTS_PATHCOVERING;
                           break;
               case 'i':   if (params.fileContents[curIndex] & FILECONTENTS_PREFIXCOVERING) {
                              strcat (error, "Duplicate use of `i' flag for file contents");
                              return false;
                           }
                           params.fileContents[curIndex] |= FILECONTENTS_PREFIXCOVERING;
                           break;
               case 'b':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_BGPDUMP;
                           typeSet = true;
                           break;
               case 'B':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_BGPDUMP;
                           expectVpoint = true;
                           typeSet = true;
                           break;
               case 's':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_SAT;
                           params.nonBGPfiles++;
                           typeSet = true;
                           break;
               case 'v':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_VANTAGE;
                           params.nonBGPfiles++;
                           typeSet = true;
                           break;
               case 'p':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_PATH;
                           params.nonBGPfiles++;
                           typeSet = true;
                           break;
               case 'u':   if (typeSet) {
                              strcat (error, "Possibly ambiguous file type: `");
                              strcat (error, argv[i]);
                              strcat (error, "'");
                              return false;
                           }
                           params.fileType[curIndex] = FILETYPE_UNKNOWN;
                           params.nonBGPfiles++;
                           typeSet = true;
                           break;
               default:    strcat (error, "Invalid file type/contents flag: `");
                           sprintf(error, "%s%c'", error, argv[i][i2]);
                           return false;
                           break;
            }
         }
         if (!typeSet || params.fileContents[curIndex] == 0) {
            strcat (error, "At least one file type and one file content are needed");
            return false;
         }
         if (!expectVpoint) {
            expectFile = true;
         }
         parsed = true;
      }
      if (!parsed) {
         if (params.outFile == NULL) {
            // This is the output file name
            params.outFile = new char[strlen(argv[i]) + 1];
            strcpy (params.outFile, argv[i]);
            parsed = true;
         }
         else {
            // Sorry, no other chance to parse this token
            strcat (error, "Invalid option: `");
            strcat (error, argv[i]);
            strcat (error, "'");
            return false;
         }
      }
   }
   if (expectFile) {
      strcat (error, "File name expected after --file option");
      return false;
   }
   if (expectGraphFile) {
      strcat (error, "Graph file name expected after --enhance option");
      return false;
   }
   if (expectVpoint) {
      strcat (error, "Vantage point number expected after --file option");
      return false;
   }
   if (expectComment) {
      strcat (error, "Comment expected after --comment option");
      return false;
   }
   if (expectOutFile) {
      strcat (error, "Output file name expected after -o option");
      return false;
   }
   if (!params.showHelp && (params.graphFile == NULL) &&
       !(processedContents & FILECONTENTS_NODES)) {
      strcat (error, "At least a file to get nodes from is needed");
      return false;
   }
   if (processedContents & FILECONTENTS_EDGEDIRECTION &&
       (params.graphFile == NULL) &&
       !(processedContents & FILECONTENTS_EDGES)) {
      strcat (error, "Cannot assign edge directions without a file to get edges from");
      return false;
   }
   if (processedContents & (FILECONTENTS_PATHCOVERING | FILECONTENTS_PREFIXCOVERING) &&
       (params.graphFile == NULL) &&
       !(processedContents & FILECONTENTS_EDGES)) {
      strcat (error, "Cannot use covering information without a file to get edges from");
      return false;
   }
   if (!processedContents && params.graphFile != NULL && !commentSet) {
      strcat (error, "This would not change so much in the existing graph...");
      return false;
   }

   if (!params.showHelp) {
      params.BGPfiles = curIndex - params.nonBGPfiles;
      curIndex = 0;
      /* Now sort file names according to their contents, as follows:
         nodes files
         edges files (if any)
         edge direction files (if any)
         path covering files (if any)
         IP address covering files (if any) */
      for (int content = FILECONTENTS_NODES;
           content <= FILECONTENTS_PREFIXCOVERING;
           content <<= 1) {
         for (int i = curIndex; i <= params.BGPfiles + params.nonBGPfiles - 1; i++) {
            if (params.fileContents[i] & content) {
               dynaSwap(params, i, curIndex);
               curIndex++;
            }
         }
      }
   }
   
   return true;
}

// This function just prints the name of the tool
void printHeader(FILE *file) {
   static bool alreadyPrinted = false;
   if (!alreadyPrinted) {
      if (isTerm(file)) {
         fprintf (file, "%stmakeGraph%s 1.3\n", CONSOLE_BOLD, CONSOLE_NORMAL);
      }
      else {
         fprintf (file, "tmakeGraph 1.3\n");
      }
      alreadyPrinted = true;
   }
}

/* This function checks whether there is at least a true element in
   an array of boolean elements */
bool anyTrue (bool array[], int howManyElements) {
   for (int i = 0; i <= howManyElements - 1; i++) {
      if (array[i]) {
         return true;
      }
   }
   return false;
}

// This function prints out a command line help
void printHelp(FILE *file, char *cmdLine) {
   char CON_BOLD[strlen(CONSOLE_BOLD) + 1];
   char CON_UNDERLINE[strlen(CONSOLE_UNDERLINE) + 1];
   char CON_NORMAL[strlen(CONSOLE_NORMAL) + 1];
   if (isTerm(file)) {
      strcpy (CON_BOLD, CONSOLE_BOLD);
      strcpy (CON_UNDERLINE, CONSOLE_UNDERLINE);
      strcpy (CON_NORMAL, CONSOLE_NORMAL);
   }
   else {
      strcpy (CON_BOLD, "");
      strcpy (CON_UNDERLINE, "");
      strcpy (CON_NORMAL, "");
   }
   printHeader(file);
   fprintf (file, "%sTORQUE toolkit%s AS graph builder\n", CON_BOLD, CON_NORMAL);
   fprintf (file, "USAGE: %s [options] [[-o] <outFile>] {--file:{bBsvpunedci} <filename>...}\n\n",
      cmdLine);
   fprintf (file, "%sInput files%s must be provided using --file option, in the following way:\n",
      CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  --file:  this preamble is used to indicate that a file name will follow;\n");
   fprintf (file, "           each file must be identified by its format (BGP dump, list of\n");
   fprintf (file, "           AS paths, etc.) and the contents that must be extracted from it\n");
   fprintf (file, "           (nodes, edges, etc.)\n\n");
   fprintf (file, "  {bBsvpu} each `--file:' option must be followed by exactly one of these\n");
   fprintf (file, "           letters, which identify file format as follows:\n");
   fprintf (file, "             b,B  file is a BGP dump (see below for an explanation of the\n");
   fprintf (file, "                  difference between 'b' and 'B')\n");
   fprintf (file, "             s    file is the output of a bgpPathSat computation\n");
   fprintf (file, "             v    file is a relationship file (i.e., each line looks as\n");
   fprintf (file, "                  follows: `<AS1> <AS2> <flag>'; when flag is 1, AS1 is\n");
   fprintf (file, "                  supposed to be a provider and AS2 a customer; if flag\n");
   fprintf (file, "                  is 0, AS1 and AS2 have a peering relationship)\n");
   fprintf (file, "             p    file is a plain list of AS paths (i.e., paths must\n");
   fprintf (file, "                  neither be groupped by number of occurrences nor tagged)\n");
   fprintf (file, "             u    file is a list of couples of ASs whose corresponding\n");
   fprintf (file, "                  edges could not be assigned a direction\n\n");
   fprintf (file, "  {nedci}  each `--file:' option must also be followed by at least one of\n");
   fprintf (file, "           these letters, which identify what to get from current file:\n");
   fprintf (file, "             n    get graph nodes from current file; do not use if file\n");
   fprintf (file, "                  format is 'b'\n");
   fprintf (file, "             e    get graph edges (but not their direction) from current\n");
   fprintf (file, "                  file; edges are inserted in the graph only if they do\n");
   fprintf (file, "                  not exist yet; otherwise, they are left unchanged; do\n");
   fprintf (file, "                  not use if file format is 'b'\n");
   fprintf (file, "             d    get direction of already existing edges from current\n");
   fprintf (file, "                  file; do not use if file format is either 'b' or 'u'\n");
   fprintf (file, "             c    get path covering (number of paths that traverse each\n");
   fprintf (file, "                  edge) from current file; do not use if file format is\n");
   fprintf (file, "                  any one of 'b', 'v' or 'u'\n");
   fprintf (file, "             i    get IP address covering (number of IP addresses that are\n");
   fprintf (file, "                  announced over each edge) from current file; only use if\n");
   fprintf (file, "                  file format is 'b'\n\n");
   fprintf (file, "  <filename>  the name of the current input file\n\n\n");
   fprintf (file, "Option --file can be used with more than one content per file; this allows\n");
   fprintf (file, "to optimize file access, by avoiding that each file is read more than\n");
   fprintf (file, "once; anyway, regardless of the order in which files are provided on\n");
   fprintf (file, "command line, they are always read in the following order:\n");
   fprintf (file, "nodes files, edges files, edge directions files, path covering files, IP\n");
   fprintf (file, "address covering files.\n");
   fprintf (file, "Minimal input consists of a file to get nodes from.\n\n");
   fprintf (file, "When using option --file to feed more than one file with the same\n");
   fprintf (file, "contents, the result is that such contents are merged together (e.g., all\n");
   fprintf (file, "the nodes from all the files are inserted in the graph - of course, with\n");
   fprintf (file, "no duplicates... -, all the IP addresses found in all the BGP dumps are\n");
   fprintf (file, "supposed to be exchanged over the appropriate edges - without considering\n");
   fprintf (file, "addresses more than once -, etc.).\n\n");
   fprintf (file, "File ordering is significant; in fact, if (for example) two node files A\n");
   fprintf (file, "and B are provided, and A is also an edge file, reading A before B implies\n");
   fprintf (file, "that none of the edges between nodes in B (but not in A) will be inserted.\n");
   fprintf (file, "The only exception are the BGP dumps: since they are ordered according to\n");
   fprintf (file, "the value of the network, the sequence in which they are provided on the\n");
   fprintf (file, "command line can be arbitrary.\n\n");
   fprintf (file, "If the AS paths inside the BGP dumps do not include receiving vantage\n");
   fprintf (file, "point number, it is possible to use 'B' instead of 'b'; by doing this,\n");
   fprintf (file, "following parameters are expected to be the number of the AS that should\n");
   fprintf (file, "be prepended to each path and the name of the file containing the BGP dump.\n\n");
   fprintf (file, "BGP dumps are processed as follows: no check is performed for both cycles\n");
   fprintf (file, "and private ASs; AS sets are just skipped; prepending is automatically\n");
   fprintf (file, "skipped because AS graphs (usually...) do not have self-edges; incomplete\n");
   fprintf (file, "paths are considered as they appear.\n\n\n");
   fprintf (file, "%sOptions%s are:\n", CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  --comment <string>  Insert a comment string inside the graph (e.g., to\n");
   fprintf (file, "                      make it easier to understand which data set it comes\n");
   fprintf (file, "                      from); by default, graph is also labelled with the\n");
   fprintf (file, "                      names of the files it was built from\n");
   fprintf (file, "  --enhance <filename>  Read an existing graph file and perform some sort\n");
   fprintf (file, "                        of update over it; when using this option, input\n");
   fprintf (file, "                        files can be of any type (i.e., minimal input is\n");
   fprintf (file, "                        not a nodes file any more); it is also possible to\n");
   fprintf (file, "                        give no input files if, for example, one just\n");
   fprintf (file, "                        wants to update graph comment; this is the only\n");
   fprintf (file, "                        file name that can be replaced with '-' if input\n");
   fprintf (file, "                        graph is provided on standard input\n");
   fprintf (file, "  -f      Don't prompt user (default behaviour) before overwriting\n");
   fprintf (file, "          existing files\n");
   fprintf (file, "  -v      Be verbose: report all errors that occur during file parsing\n");
   fprintf (file, "  --help  Show this help\n\n\n");
   fprintf (file, "<outFile> is the name of the output file graph. If none is given, the\n");
   fprintf (file, "graph will be dumped to stdout. Use -o option when ambiguities may arise\n");
   fprintf (file, "between file name and other command line options.\n");
}

/* This function builds an internal representation of the desired graph,
   based on an array of user-provided file names (`fileName'); please
   ensure it stores file names in the order in which they must be actually
   read (i.e., the first file is the one to get nodes from, etc.; BGP
   dump files must be the last ones); returned value is 0 upon success;
   non-zero if a fail occurs */
int buildGraph(short int nonBGPfiles, short int BGPfiles, char **fileName,
               file_type *fileType, int *fileContents,
               bool showPercentage, bool verbose, ASnode *prependAS, ASGraph &graph) {
   Parser *parser;
   pathParser::parsed_data *pathParsedData;
   satParser::parsed_data *satParsedData;
   vantageParser::parsed_data *vantageParsedData;
   unknownParser::parsed_data *unknownParsedData;
   bgpParser bgp_parser[BGPfiles];

   List<unsigned short int> *tempASpath1;
   List<bgpParser::ASItem> *tempASpath2;

   ASnode lastNode;
   IPPrefix curPrefix[BGPfiles];
   List<ASnode> nodeList2;
   ASEdgeWeight zeroWeight;
   ASGraph::graph_edge *curEdge, *existingEdge;

   bool goOn, notFirst, rev, firstLine,
        continueParsing[BGPfiles];
   short int i, minIndex;
   float readBytes, totBytes;
   unsigned long long int newAddresses;

   for (i = 0; i <= nonBGPfiles - 1; i++) {
      switch (fileType[i]) {
         case FILETYPE_BGPDUMP:  parser = new bgpParser;
                                 break;
         case FILETYPE_SAT:      parser = new satParser;
                                 break;
         case FILETYPE_VANTAGE:  parser = new vantageParser;
                                 break;
         case FILETYPE_PATH:     parser = new pathParser;
                                 break;
         case FILETYPE_UNKNOWN:  parser = new unknownParser;
                                 break;
         default:                break;
      }

      // Try to open current input file
      if (!parser->openFile(fileName[i])) {
         printHeader (stderr);
         fprintf (stderr, "Error: Unable to open input file `%s'\n", fileName[i]);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return 2;
      }

      pathParsedData = &(((pathParser*) parser)->parsedData);
      satParsedData = &(((satParser*) parser)->parsedData);
      vantageParsedData = &(((vantageParser*) parser)->parsedData);
      unknownParsedData = &(((unknownParser*) parser)->parsedData);

      if (showPercentage) {
         notFirst = false;
         printf ("Reading");
         if (fileContents[i] & FILECONTENTS_NODES) {
            printf (" nodes");
            notFirst = true;
         }
         if (fileContents[i] & FILECONTENTS_EDGES) {
            if (notFirst) {
               printf (",");
            }
            printf (" edges");
            notFirst = true;
         }
         if (fileContents[i] & FILECONTENTS_EDGEDIRECTION) {
            if (notFirst) {
               printf (",");
            }
            printf (" edge directions");
            notFirst = true;
         }
         if (fileContents[i] & FILECONTENTS_PATHCOVERING) {
            if (notFirst) {
               printf (",");
            }
            printf (" path covering");
            notFirst = true;
         }
         printf (" file (step %i of %i)\n", i + 1, nonBGPfiles + (BGPfiles > 0 ? 1 : 0));
         printf ("`%s'\n", fileName[i]);
      }

      // Parse current input file
      while (parser->nextLine() == PARSEFILE_NOERROR) {
         if (parser->parseLine() != PARSE_NOERROR) {
            // Parse error
            if (verbose) {
               fprintf (stderr, "\n");
               printHeader(stderr);
               fprintf (stderr, "\nParse error in line %lu of file `%s':\n",
                  parser->readLines(), fileName[i]);
               fprintf (stderr, "%s\n", parser->currentLine());
               fprintf (stderr, "%s\n", parser->errReason());
            }
         }
         else {
            if (showPercentage) {
               // Print percentage of completion
               printPercentage ((float) parser->filePos(), (float) parser->fileSize());
            }

            if (fileType[i] == FILETYPE_PATH) {
               tempASpath1 = &(pathParsedData->ASpath);
            }

            if (fileContents[i] & FILECONTENTS_NODES) {
               // Retrieve nodes from this file
               switch (fileType[i]) {
                  case FILETYPE_PATH:
                     // Get nodes from the entire AS path
                     goOn = !tempASpath1->isEmpty();
                     for (tempASpath1->goToFirst(); goOn; goOn = tempASpath1->next()) {
                        graph.addNode(tempASpath1->current(),
                                      tempASpath1->current());
                     }
                     break;
                  case FILETYPE_SAT:
                     // Get the two nodes in current line
                     if (!satParsedData->peering && !satParsedData->bad) {
                        graph.addNode (satParsedData->from, satParsedData->from);
                        graph.addNode (satParsedData->to, satParsedData->to);
                     }
                     break;
                  case FILETYPE_VANTAGE:
                     // Get the two nodes in current line
                     graph.addNode (vantageParsedData->from,
                                    vantageParsedData->from);
                     graph.addNode (vantageParsedData->to,
                                    vantageParsedData->to);
                     break;
                  case FILETYPE_UNKNOWN:
                     // Get the two nodes in current line
                     graph.addNode (unknownParsedData->from,
                                    unknownParsedData->from);
                     graph.addNode (unknownParsedData->to,
                                    unknownParsedData->to);
                     break;
                  default:
                     // This can never happen (no other file can contain nodes)
                     break;
               }
            }
            if (fileContents[i] & FILECONTENTS_EDGES) {
               // Retrieve edges from this file
               switch (fileType[i]) {
                  case FILETYPE_PATH:
                     // Get edges from the entire AS path
                     goOn = tempASpath1->len() > 1;
                     lastNode = tempASpath1->first();
                     tempASpath1->goToFirst();
                     for (tempASpath1->next(); goOn; goOn = tempASpath1->next()) {
                        existingEdge = graph.edgeProperties(lastNode, tempASpath1->current());
                        if (existingEdge == NULL) {
                           /* Edge does not exist yet (the above check must
                              be performed because when trying to insert
                              an existing edge, its properties are cleared */
                           graph.addEdge(lastNode,
                                         tempASpath1->current(),
                                         DIR_NONE,
                                         zeroWeight,
                                         zeroWeight);
                        }
                        else {
                           delete existingEdge->from;
                           delete existingEdge->to;
                           delete existingEdge;
                        }
                        lastNode = tempASpath1->current();
                     }
                     break;
                  case FILETYPE_SAT:
                     // Get the only edge in current line
                     if (!satParsedData->peering && !satParsedData->bad) {
                        existingEdge = graph.edgeProperties(satParsedData->from,
                            satParsedData->to);
                        if (existingEdge == NULL) {
                           /* Edge does not exist yet (the above check must
                              be performed because when trying to insert
                              an existing edge, its properties are cleared */
                           graph.addEdge (satParsedData->from,
                                          satParsedData->to,
                                          DIR_NONE,
                                          zeroWeight,
                                          zeroWeight);
                        }
                        else {
                           delete existingEdge->from;
                           delete existingEdge->to;
                           delete existingEdge;
                        }
                     }
                     break;
                  case FILETYPE_VANTAGE:
                     // Get the only edge in current line
                     existingEdge = graph.edgeProperties(vantageParsedData->from,
                         vantageParsedData->to);
                     if (existingEdge == NULL) {
                        /* Edge does not exist yet (the above check must
                           be performed because when trying to insert
                           an existing edge, its properties are cleared */
                        graph.addEdge (vantageParsedData->from,
                                       vantageParsedData->to,
                                       DIR_NONE,
                                       zeroWeight,
                                       zeroWeight);
                     }
                     else {
                        delete existingEdge->from;
                        delete existingEdge->to;
                        delete existingEdge;
                     }
                     break;
                  case FILETYPE_UNKNOWN:
                     // Get the only edge in current line
                     existingEdge = graph.edgeProperties(unknownParsedData->from,
                         unknownParsedData->to);
                     if (existingEdge == NULL) {
                        /* Edge does not exist yet (the above check must
                           be performed because when trying to insert
                           an existing edge, its properties are cleared */
                        graph.addEdge (unknownParsedData->from,
                                       unknownParsedData->to,
                                       DIR_NONE,
                                       zeroWeight,
                                       zeroWeight);
                     }
                     else {
                        delete existingEdge->from;
                        delete existingEdge->to;
                        delete existingEdge;
                     }
                     break;
                  default:
                     // This can never happen (no other file can contain edges)
                     break;
               }
            }
            if (fileContents[i] & FILECONTENTS_EDGEDIRECTION) {
               // Retrieve edge directions from this file
               switch (fileType[i]) {
                  case FILETYPE_PATH:
                     /* Assume edge direction is given by the order of
                        ASs in the AS path (i.e., X Y means that there will be
                        an edge from X to Y) */
                     goOn = tempASpath1->len() > 1;
                     lastNode = tempASpath1->first();
                     tempASpath1->goToFirst();
                     for (tempASpath1->next(); goOn; goOn = tempASpath1->next()) {
                        curEdge = graph.edgeProperties(lastNode, tempASpath1->current());
                        if (curEdge != NULL) {
                           // Edge exists in the graph
                           // Check whether a direction has already been assigned
                           if (!curEdge->weight_from->oriented) {
                              *(curEdge->orientation) = DIR_OUT;
                              curEdge->weight_from->oriented = true;
                              curEdge->weight_to->oriented = true;
                           }
                           else {                              
                              if (*(curEdge->orientation) == DIR_IN)
                                 *(curEdge->orientation) = DIR_NONE;
                              else if (*(curEdge->orientation) != DIR_NONE)
                                 *(curEdge->orientation) = DIR_OUT;
                           }
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                           curEdge = graph.edgeProperties(tempASpath1->current(), lastNode);
                           if (!curEdge->weight_from->oriented) {
                              *(curEdge->orientation) = DIR_IN;
                              curEdge->weight_from->oriented = true;
                              curEdge->weight_to->oriented = true;
                           }
                           else {
                              if (*(curEdge->orientation) == DIR_OUT)
                                 *(curEdge->orientation) = DIR_NONE;
                              else if (*(curEdge->orientation) != DIR_NONE)
                                 *(curEdge->orientation) = DIR_IN;
                           }
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                        }
                        lastNode = tempASpath1->current();
                     }
                     break;
                  case FILETYPE_SAT:
                     // Get the direction of the only edge in current line
                     if (!satParsedData->peering && !satParsedData->bad) {
                        curEdge = graph.edgeProperties(
                                    satParsedData->from,
                                    satParsedData->to);
                        if (curEdge != NULL) {
                           // Edge exists in the graph
                           *(curEdge->orientation) = DIR_OUT;
                           curEdge->weight_from->oriented = true;
                           curEdge->weight_to->oriented = true;
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                           curEdge = graph.edgeProperties(
                                       satParsedData->to,
                                       satParsedData->from);
                           *(curEdge->orientation) = DIR_IN;
                           curEdge->weight_from->oriented = true;
                           curEdge->weight_to->oriented = true;
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                        }
                     }
                     break;
                  case FILETYPE_VANTAGE:
                     // Get the direction of the only edge in current line
                     curEdge = graph.edgeProperties(
                                 vantageParsedData->from,
                                 vantageParsedData->to);
                     if (curEdge != NULL) {
                        // Edge exists in the graph
                        *(curEdge->orientation) = (vantageParsedData->peering ?
                                                   DIR_NONE : DIR_OUT);
                        curEdge->weight_from->oriented = true;
                        curEdge->weight_to->oriented = true;
                        delete curEdge->from;
                        delete curEdge->to;
                        delete curEdge;
                        curEdge = graph.edgeProperties(
                                    vantageParsedData->to,
                                    vantageParsedData->from);
                        *(curEdge->orientation) = (vantageParsedData->peering ?
                                                   DIR_NONE : DIR_IN);
                        curEdge->weight_from->oriented = true;
                        curEdge->weight_to->oriented = true;
                        delete curEdge->from;
                        delete curEdge->to;
                        delete curEdge;
                     }
                     break;
                  default:
                     // This can never happen (no other file can contain edge
                     // directions)
                     break;
               }
            }
            if (fileContents[i] & FILECONTENTS_PATHCOVERING) {
               // Retrieve information about path covering from this file
               switch (fileType[i]) {
                  case FILETYPE_PATH:
                     /* Each edge stores in two different locations the number of paths that
                        traversed it in either direction */
                     goOn = tempASpath1->len() > 1;
                     lastNode = tempASpath1->first();
                     tempASpath1->goToFirst();
                     for (tempASpath1->next(); goOn; goOn = tempASpath1->next()) {
                        curEdge = graph.edgeProperties(lastNode, tempASpath1->current());
                        if (curEdge != NULL) {
                           // Edge exists in the graph
                           curEdge->weight_from->pathCovering[CMPGRAPH_FIRST]++;
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                           curEdge = graph.edgeProperties(tempASpath1->current(), lastNode);
                           curEdge->weight_to->pathCovering[CMPGRAPH_FIRST]++;
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                        }
                        lastNode = tempASpath1->current();
                     }
                     break;
                  case FILETYPE_SAT:
                     /* SAT files just contain numerical information about path
                        covering, so all that is to be done is getting them */
                     if (!satParsedData->peering && !satParsedData->bad &&
                        satParsedData->straight != NULL && satParsedData->negated != NULL) {
                        curEdge = graph.edgeProperties(
                                    satParsedData->from,
                                    satParsedData->to);
                        if (curEdge != NULL) {
                           // Edge exists in the graph
                           rev = satParsedData->reversed;
                           curEdge->weight_from->pathCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->straight) :
                                                         *(satParsedData->negated));
                           curEdge->weight_to->pathCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->negated) :
                                                         *(satParsedData->straight));
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                           curEdge = graph.edgeProperties(
                                    satParsedData->to,
                                    satParsedData->from);
                           curEdge->weight_from->pathCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->negated) :
                                                         *(satParsedData->straight));
                           curEdge->weight_to->pathCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->straight) :
                                                         *(satParsedData->negated));
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                        }
                     }
                     break;
                  default:
                     // This can never happen (no other file can contain path
                     // covering information)
                     break;
               }
            }
            if (fileContents[i] & FILECONTENTS_PREFIXCOVERING) {
               // Retrieve information about IP address covering from this file
               switch (fileType[i]) {
                  case FILETYPE_SAT:
                     /* SAT files just contain numerical information about IP address
                        covering, so all that is to be done is getting them */
                     if (!satParsedData->peering && !satParsedData->bad &&
                        satParsedData->straight != NULL && satParsedData->negated != NULL) {
                        curEdge = graph.edgeProperties(
                                    satParsedData->from,
                                    satParsedData->to);
                        if (curEdge != NULL) {
                           // Edge exists in the graph
                           rev = satParsedData->reversed;
                           curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->straight) :
                                                         *(satParsedData->negated));
                           curEdge->weight_to->IPCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->negated) :
                                                         *(satParsedData->straight));
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                           curEdge = graph.edgeProperties(
                                    satParsedData->to,
                                    satParsedData->from);
                           curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->negated) :
                                                         *(satParsedData->straight));
                           curEdge->weight_to->IPCovering[CMPGRAPH_FIRST] +=
                              (unsigned long int) (rev ? *(satParsedData->straight) :
                                                         *(satParsedData->negated));
                           delete curEdge->from;
                           delete curEdge->to;
                           delete curEdge;
                        }
                     }
                     break;
                  default:
                     // This can never happen (no other file can contain prefix
                     // covering information)
                     break;
               }
            }
         }
      }

      if (showPercentage) {
         // Show 100%
         printPercentage ((float) parser->filePos(), (float) parser->fileSize());
         printf ("\n");
      }

      parser->closeFile();
      delete parser;
   }

   // Now get IP address covering from BGP dump (if requested)
   if (showPercentage && BGPfiles > 0) {
      printf ("Reading IP covering file(s) (step %i of %i)\n",
         nonBGPfiles + 1, nonBGPfiles + 1);
   }
   for (i = 0; i <= BGPfiles - 1; i++) {
      // Try to open current input file
      continueParsing[i] = false;
      if (!bgp_parser[i].openFile(fileName[i + nonBGPfiles])) {
         printHeader (stderr);
         fprintf (stderr, "Error: Unable to open input file `%s'\n", fileName[i + nonBGPfiles]);
         fprintf (stderr, "(%s)\n", strerror(errno));
         // Close successfully opened files
         for (short int i2 = 0; i2 <= i - 1; i++) {
            bgp_parser[i].closeFile();
         }
         return 2;
      }
      if (showPercentage) {
         printf ("`%s'\n", fileName[i + nonBGPfiles]);
      }
   }

   // Go to first line on every file
   totBytes = 0;
   for (i = 0; i <= BGPfiles - 1; i++) {
      continueParsing[i] = (bgp_parser[i].nextLine() == PARSEFILE_NOERROR);
      totBytes += bgp_parser[i].fileSize();
   }
   firstLine = true;
   while (anyTrue(continueParsing, BGPfiles)) {
      readBytes = 0;
      for (i = 0; i <= BGPfiles - 1; i++) {
         readBytes += bgp_parser[i].filePos();
         if (continueParsing[i]) {
            goOn = bgp_parser[i].parseLine() == PARSE_NOERROR;
            while (!goOn) {
               // Parse error: look for a following parsable line
               if (verbose) {
                  fprintf (stderr, "\n");
                  printHeader(stderr);
                  fprintf (stderr, "Parse error in line %lu of file `%s':\n",
                     bgp_parser[i].readLines(), fileName[i + nonBGPfiles]);
                  fprintf (stderr, "%s\n", bgp_parser[i].currentLine());
                  fprintf (stderr, "%s\n", bgp_parser[i].errReason());
               }
               continueParsing[i] = (bgp_parser[i].nextLine() == PARSEFILE_NOERROR);
               goOn = !continueParsing[i] || (bgp_parser[i].parseLine() == PARSE_NOERROR);
            }
         }
      }
      if (showPercentage) {
         // Show percentage of completion
         printPercentage (readBytes, totBytes);
      }

      /* Retrieve "minimum" prefix among those in current line of
         each BGP dump */
      minIndex = -1;                // Dummy temporary index
      for (i = 0; i <= BGPfiles - 1; i++) {
         if (bgp_parser[i].parsedData.netmask == NULL) {
            curPrefix[i] = IPPrefix(bgp_parser[i].parsedData.network);
         }
         else {
            curPrefix[i] = IPPrefix(bgp_parser[i].parsedData.network,
                                    *(bgp_parser[i].parsedData.netmask));
         }
         if (continueParsing[i]) {
            if (minIndex == -1 ||
                (curPrefix[i].address() < curPrefix[minIndex].address() ||
                (curPrefix[i].address() == curPrefix[minIndex].address() &&
                 curPrefix[i].netmask() > curPrefix[minIndex].netmask()))) {
               /* A prefix is supposed to be smaller than another if
                  its IP address is smaller or, being equal the IP addresses,
                  its netmask is larger (more '1's) */
               minIndex = i;
            }
         }
      }
      // Advance on file that owns "minimum" prefix
      continueParsing[minIndex] = (bgp_parser[minIndex].nextLine() == PARSEFILE_NOERROR);

      // Update IP covering
      tempASpath2 = &(bgp_parser[minIndex].parsedData.ASpath);
      goOn = tempASpath2->len() > 1;
      tempASpath2->goToFirst();
      if (prependAS[minIndex + nonBGPfiles] > 0 &&
          prependAS[minIndex + nonBGPfiles] != tempASpath2->current().ASnum) {
         lastNode = prependAS[minIndex + nonBGPfiles];
      }
      else {
         lastNode = tempASpath2->first().ASnum;
         tempASpath2->next();
      }
      for (; goOn; goOn = tempASpath2->next()) {
         if (!(tempASpath2->current().flags & AS_SET)) {
            // Current AS is not inside an AS set
            curEdge = graph.edgeProperties(lastNode, tempASpath2->current().ASnum);
            if (curEdge != NULL) {
               // Edge exists in the graph: update it with current prefix
               /* Update must be performed in the following way:
                  if edge is being traversed by a set X of addresses, its IP
                  covering must be increased by a value that is the number
                  of addresses in X that are new with respect to the last
                  set of addresses that traversed this edge; this is possible
                  since prefixes in a BGP dump are ordered according to address
                  (from lower to higher) and netmask strictness (from more
                  strict to less strict) */
               if (curEdge->weight_to->IPCovering[CMPGRAPH_FIRST] == 0) {
                  /* This is the first time this edge is being traversed by
                     IP addresses (i.e., there is no previous set of addresses
                     that traversed this edge) */
                  curEdge->weight_to->IPCovering[CMPGRAPH_FIRST] =
                     curPrefix[minIndex].addresses();
               }
               else {
                  newAddresses = curPrefix[minIndex].addresses() -
                     IPSubSpace(curPrefix[minIndex]).
                        intersect(curEdge->weight_to->lastPrefix).addresses();
                  curEdge->weight_to->IPCovering[CMPGRAPH_FIRST] += newAddresses;
               }
               curEdge->weight_to->lastPrefix = curPrefix[minIndex];
               delete curEdge->from;
               delete curEdge->to;
               delete curEdge;
               curEdge = graph.edgeProperties(tempASpath2->current().ASnum, lastNode);
               if (curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] == 0) {
                  /* This is the first time this edge is being traversed by
                     IP addresses (i.e., there is no previous set of addresses
                     that traversed this edge) */
                  curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] =
                     curPrefix[minIndex].addresses();
               }
               else {
                  curEdge->weight_from->IPCovering[CMPGRAPH_FIRST] += newAddresses;
               }
               curEdge->weight_from->lastPrefix = curPrefix[minIndex];
               delete curEdge->from;
               delete curEdge->to;
               delete curEdge;
            }
            lastNode = tempASpath2->current().ASnum;
         }
      }
   }
   // Show 100%
   if (BGPfiles > 0 && showPercentage) {
      printPercentage (totBytes, totBytes);
      printf ("\n");
   }
   for (i = 0; i <= BGPfiles - 1; i++) {
      bgp_parser[i].closeFile();
   }


   return 0;
}

int main(int argc, char **argv) {
   ASGraph graph;
   char error[513];           // One more character for '\0'
   int retValue;
   cmdLineParams params;
   bool stdInIsTerm, stdOutIsTerm;
   FILE *graphFile;
   
   if (!parseArguments(argc, argv, params, error)) {
      printHeader(stderr);
      fprintf (stderr, "Syntax error: %s\n", error);
      fprintf (stderr, "Type %s --help for more information\n", argv[0]);
      return 1;
   }

   // Test whether standard streams are connected to a terminal
   stdInIsTerm = isTerm(stdin);
   stdOutIsTerm = isTerm(stdout);

   if (params.showHelp) {
      printHelp (stdout, argv[0]);
      return 0;
   }

   if (stdOutIsTerm) {
      printHeader(stdout);
   }

   /* Avoid asking user for file overwriting only after
      processing */
   if (params.outFile != NULL && !params.force) {
      graphFile = fopen(params.outFile, "r");
      if (graphFile != NULL) {
         fclose(graphFile);
         if (!askOverwrite(params.outFile)) {
            return 3;
         }
      }
   }

   if (params.graphFile != NULL) {
      // Read source graph...
      if (strcmp(params.graphFile, "-") == 0) {
         // ...from stdin
         if (!graph.readFromFile(stdin)) {
            printHeader(stderr);
            fprintf (stderr, "Input does not appear to be a valid graph:\n");
            fprintf (stderr, "%s\n", graph.errReason());
            return 3;
         }
      }
      else {
         graphFile = fopen(params.graphFile, "r");
         if (graphFile == NULL) {
            printHeader(stderr);
            fprintf (stderr, "Error while opening source graph `%s'\n",
               params.graphFile);
            fprintf (stderr, "(%s)\n", strerror(errno));
            return 3;
         }
         else {
            if (stdOutIsTerm) {
               printf ("Reading input graph file...\n");
            }
            if (!graph.readFromFile(graphFile)) {
               printHeader(stderr);
               fprintf (stderr, "File `%s' does not appear to be a valid graph file:\n",
                  params.graphFile);
               fprintf (stderr, "%s\n", graph.errReason());
               return 3;
            }
         }
      }
   }

   retValue = buildGraph (params.nonBGPfiles, params.BGPfiles, params.fileName,
                          params.fileType, params.fileContents,
                          stdOutIsTerm && params.outFile != NULL,
                          params.verbose, params.prependAS, graph);

   if (retValue != 0) {
      return retValue;
   }

   graph.setSourceFiles(params.fileName, false);
   graph.setComment(params.comment);

   if (params.outFile == NULL) {
      // Dump graph to stdout
      if (!graph.writeToFile(stdout)) {
         printHeader(stderr);
         fprintf (stderr, "Error while dumping graph:\n");
         fprintf (stderr, "%s\n", graph.errReason());
         return 3;
      }
   }
   else {
      // Save graph to file
      graphFile = fopen(params.outFile, "w");
      if (graphFile != NULL) {
         if (stdOutIsTerm) {
            printf ("Saving graph to file...\n");
         }
         if (!graph.writeToFile(graphFile)) {
            printHeader(stderr);
            fprintf (stderr, "Error while writing graph to file `%s':\n", params.outFile);
            fprintf (stderr, "%s\n", graph.errReason());
            fclose(graphFile);
            return 3;
         }
         fclose(graphFile);
      }
      else {
         printHeader(stderr);
         fprintf (stderr, "Unable to open file `%s' for writing\n", params.outFile);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return 2;
      }
   }

   return 0;
}
